@@ -1,5 +1,5 @@
-__version__ = "0.5.7"
-__description__ = "Calibrate the internal and external parameters of cameras, rectify stereo cameras by OpenCV python."
+__version__ = "0.5.13"
+__description__ = "Calibrate camera's intrinsic/extristric, and build stereo depth camera with OpenCV python."
 __license__ = "MIT"
 __author__ = "DIYer22"
 __author_email__ = "ylxx@live.com"
