@@ -7,13 +7,14 @@ import os
 import cv2
 import yaml
 import copy
+import uuid
 import pickle
 import numpy as np
 from glob import glob
 from tqdm import tqdm
 
 with boxx.inpkg():
-    from .feature_libs import CheckboardFeatureLib
+    from .feature_libs import CheckboardFeatureLib, MetaFeatureLib
     from .stereo import Stereo
     from .utils import r_t_to_T, intrinsic_format_conversion
     from . import utils
@@ -33,37 +34,39 @@ class Cam(dict):
         enable_cache=False,
     ):
         super().__init__()
+        self.img_paths = img_paths
+        self.feature_lib = feature_lib
+        self.save_feature_vis = save_feature_vis
+        self.name = name or ("cam" + str(boxx.increase("caibrating.cam-name")))
+        self.undistorted = undistorted
+        self.enable_cache = enable_cache
+
         if img_paths is None:
             return
+        # papre name to dict for calibrate
         assert len(img_paths), img_paths
         assert len(img_paths) == len(set(img_paths))
-        self.feature_lib = feature_lib
         if isinstance(img_paths, (list, tuple)):
             left_idx, right_idx = self.get_key_idx(img_paths)
             img_paths = {path[left_idx:right_idx]: path for path in sorted(img_paths)}
-
         self.img_paths = img_paths
-        self.name = name or ("cam" + str(boxx.increase("caibrating.cam-name")))
-
-        self.enable_cache = enable_cache
         if self._cache():
             return
         for key in tqdm(sorted(img_paths)):
             path = img_paths[key]
-            d = {}
-            d["path"] = path
-
-            img = boxx.imread(path)
-            d["img"] = self.process_img(img)
+            d = dict(path=path, img=self.process_img(imread(path)))
             feature_lib.find_image_points(d)
             d.pop("img")
             self[key] = d
+        self.calibrate()
 
-        self.image_points = self._get_points_for_cv2("image_points")
+    def calibrate(self):
+        self.set_xy()
+        assert len(self.valid_keys), "No any valid image!"
         self.object_points = self._get_points_for_cv2("object_points")
-
+        self.image_points = self._get_points_for_cv2("image_points")
         flags = 0
-        if undistorted:
+        if self.undistorted:
             flags = cv2.CALIB_ZERO_TANGENT_DIST + sum(
                 [
                     cv2.CALIB_FIX_K1,
@@ -74,27 +77,68 @@ class Cam(dict):
                     cv2.CALIB_FIX_K6,
                 ]
             )
-
+        init_K, init_D = None, None
+        # flags=cv2.CALIB_FIX_ASPECT_RATIO#+cv2.CALIB_FIX_PRINCIPAL_POINT
         self.retval, self.K, self.D, rvecs, tvecs = cv2.calibrateCamera(
-            self.object_points, self.image_points, self.xy, None, None, flags=flags,
+            self.object_points, self.image_points, self.xy, init_K, init_D, flags=flags,
         )
 
         for idx, key in enumerate(self.valid_keys):
             d = self[key]
             d["T"] = r_t_to_T(rvecs[idx], tvecs[idx])
 
-        self._cache(do_cache=True)
-        if save_feature_vis:
+        self._cache(do_cache=self.enable_cache)
+        if self.save_feature_vis:
             visdir = TEMP + "/calibrating-vis-" + self.name
             print("\nSave visualization of feature points in dir:", visdir)
             os.makedirs(visdir, exist_ok=True)
             for key in tqdm(self.valid_keys):
                 d = self[key]
                 d["img"] = self.process_img(imread(d["path"]))
-                vis = feature_lib.vis(d, self)
+                vis = d.get("feature_lib", self.feature_lib).vis(d, self)
                 boxx.imsave(visdir + "/" + key + ".jpg", vis)
                 d.pop("img")
 
+    def fine_tuning_intrinsic(self, base_cam, momenta=0.5):
+        """
+        Fine tuning intrinsic from base_cam's K and D. 
+        By add new image_points and object_points that fit base_cam's intrinsic
+        Then run self.calibrate() again
+
+        Parameters
+        ----------
+        base_cam : Cam
+            base cam.
+        momenta : float, [0~1]
+            the momenta of base_cam's intrinsic. The default is .5.
+        """
+        if momenta in (0, 1):
+            if momenta == 1:
+                self.load(base_cam.dump())
+            return self
+        d0 = self[list(self)[0]]
+        meta_feature_lib = MetaFeatureLib()
+        points = self._get_points_for_cv2()
+        newn = sum(map(len, points))
+        needn = int(newn * momenta / (1 - momenta))
+        uvs = np.random.rand(needn, 2) * [list(self.xy)]
+
+        normalized_undistort_points = cv2.undistortPoints(
+            uvs[:, None], base_cam.K, base_cam.D
+        )[:, 0]
+        xyzs = np.zeros_like(normalized_undistort_points, shape=(needn, 3))
+        xyzs[:, :2] = normalized_undistort_points
+
+        dn = int(len(points) * momenta / (1 - momenta))
+        for d_idx in range(dn):
+            d = dict(feature_lib=meta_feature_lib)
+            d["path"] = d0["path"]
+            d["image_points"] = np.float32(uvs[d_idx::dn])
+            d["object_points"] = np.float32(xyzs[d_idx::dn])
+            self[f"~base_cam_points-{self.name}-{str(uuid.uuid1())}"] = d
+        self.calibrate()
+        return self
+
     @property
     def valid_keys(self):
         return set([key for key in self if len(self[key].get("image_points", {}))])
@@ -126,17 +170,22 @@ class Cam(dict):
         rotate90 img if xy != self.xy
         """
         xy = img.shape[1], img.shape[0]
-        if not hasattr(self, "xy"):
-            self.xy = xy
+        self.set_xy(img)
+        if xy == self.xy:
+            pass
+        elif xy[::-1] == self.xy:
+            img = np.ascontiguousarray(np.rot90(img))
         else:
-            if xy == self.xy:
-                pass
-            elif xy[::-1] == self.xy:
-                img = np.ascontiguousarray(np.rot90(img))
-            else:
-                assert set(xy) == set(self.xy), f"{xy} != {self.xy}"
+            assert set(xy) == set(self.xy), f"{xy} != {self.xy}"
         return img
 
+    def set_xy(self, img=None):
+        if not hasattr(self, "xy"):
+            if img is None:
+                img = imread(next(iter(self.values()))["path"])
+            self.xy = img.shape[1], img.shape[0]
+        return self.xy
+
     def stereo_with(caml, camr):
         return Stereo(caml, camr)
 
@@ -183,7 +232,14 @@ class Cam(dict):
             rvec, tvec = np.zeros((3, 1)), np.zeros((3, 1))
         else:
             rvec, tvec = utils.T_to_r_t(T)
-        return cv2.projectPoints(xyzs, rvec, tvec, self.K, self.D)[0]
+        return cv2.projectPoints(xyzs, rvec, tvec, self.K, self.D)[0][:, 0]
+
+    def undistort_points(self, uvs):
+        K = self.K
+        if uvs.ndim == 2:
+            uvs = uvs[:, None]
+        normalized_undistort_points = cv2.undistortPoints(uvs, K, self.D)[:, 0]
+        return normalized_undistort_points * [[K[0, 0], K[1, 1]]] + [K[:2, 2]]
 
     def get_T_cam2_in_self(cam1, cam2):
         Ts = []
@@ -199,7 +255,8 @@ class Cam(dict):
             T = cam1.get_T_cam2_in_self(cam2)
         if interpolation:
             interpolation_rate = cam1.K[0, 0] / cam2.K[0, 0] * interpolation
-            interpolation_rate = max(interpolation_rate, 1)
+            if interpolation >= 1:
+                interpolation_rate = max(interpolation_rate, 1)
         else:
             interpolation_rate = 1
         point_cloud2 = utils.depth_to_point_cloud(
@@ -233,32 +290,35 @@ class Cam(dict):
         return vis
 
     def _get_points_for_cv2(self, name="image_points"):
-        points = [self[key][name] for key in self.valid_keys]
-        if isinstance(points[0], dict):
-            points = [
-                np.concatenate(
-                    [id_to_points[id] for id in sorted(id_to_points)], axis=0
-                )
-                for id_to_points in points
-            ]
+        points = []
+        for key in self.valid_keys:
+            point = self[key][name]
+            if isinstance(point, dict):
+                point = np.concatenate([point[id] for id in sorted(point)], axis=0)
+            points.append(point)
         return points
 
     def valid_keys_intersection(cam1, cam2):
         return sorted(cam1.valid_keys.intersection(cam2.valid_keys))
 
     def __str__(self,):
+        def with_delta(v, base):
+            sub = v - base
+            s = f"{round(v, 1)} Δ{'' if sub<0 else '+'}{round(sub, 1)}({round(abs(sub)/base*100, 1)}%)"
+            return s
+
         s = (
-            "Cam: \n\tname: '%s' \n\txy: %s \n\tfovs: %s \n\tK: %s \n\tD: %s \n\tvalid=%s/%s \n\tretval=%.2f \n\timage_path=%s \n"
+            "Cam(name='%s'): \n\txy: %s \n\tfovs: %s \n\tK: %s\n\tD: %s \n\tvalid=%s/%s \n\tretval=%f \n\timage_path=%s \n"
             % (
                 self.name,
                 self.xy,
                 utils._str_angle_dic(self.fovs),
-                "fx=%s, fy=%s, cx=%s, cy=%s"
+                "\n\t\tfx=%s, fy=%s"
+                % (round(self.fx, 1), with_delta(self.fy, self.fx),)
+                + "\n\t\tcx=%s, cy=%s"
                 % (
-                    round(self.fx, 1),
-                    round(self.fy, 1),
-                    round(self.cx, 1),
-                    round(self.cy, 1),
+                    with_delta(self.cx, self.xy[0] / 2),
+                    with_delta(self.cy, self.xy[1] / 2),
                 ),
                 str(self.D.round(2)),
                 len(self.__dict__.get("image_points", [])),
@@ -276,7 +336,7 @@ class Cam(dict):
         dic = {
             k: v.tolist() if isinstance(v, np.ndarray) else v
             for k, v in self.__dict__.items()
-            if k in ["D", "xy", "name", "T_in_main_cam"]
+            if k in ["D", "xy", "name", "T_in_main_cam", "retval"]
         }
         dic.update(intrinsic_format_conversion(self.K))
         if return_dict:
@@ -302,6 +362,7 @@ class Cam(dict):
         else:
             dic = copy.deepcopy(path_or_str_or_dict)
         dic["K"] = intrinsic_format_conversion(dic)
+        [dic.pop(k) for k in ("fx", "fy", "cx", "cy")]
         dic["D"] = np.float64(dic["D"])
         dic["xy"] = tuple(dic["xy"])
         self.__dict__.update(dic)
@@ -312,6 +373,17 @@ class Cam(dict):
         new.load(self.dump())
         return new
 
+    def rotate(self, k=1):
+        # TODO rotate consider cam.D
+        # assert cam.dump()==cam.rotate(4).dump()
+        dic = self.dump(return_dict=True)
+        for idx in range(k):
+            dic["xy"] = dic["xy"][::-1]
+            dic["fx"], dic["fy"] = dic["fy"], dic["fx"]
+            dic["cx"], dic["cy"] = dic["xy"][0] - dic["cy"], dic["cx"]
+        new = type(self).load(dic)
+        return new
+
     @property
     def fx(self):
         return self.K[0, 0]
@@ -337,19 +409,10 @@ class Cam(dict):
         )
         return dict(fov=fov, fovx=fovx, fovy=fovy)
 
-    @classmethod
-    def init_by_K_D(cls, K, D, xy, name=None):
-        self = cls()
-        self.name = name or ("cam" + str(boxx.increase("caibrating.cam-name")))
-        self.K = K
-        self.D = D
-        self.xy = xy
-        return self
-
     @staticmethod
     def get_key_idx(paths):
         if len(paths) <= 1:
-            return 0, 1
+            return -3, None
         left_idx = 0
         while all([paths[0][left_idx] == path[left_idx] for path in paths]):
             left_idx += 1
@@ -402,6 +465,31 @@ class Cam(dict):
             )
         )
 
+    @classmethod
+    def init_by_K_D(cls, K, D, xy, name=None):
+        self = cls()
+        self.name = name or ("cam" + str(boxx.increase("caibrating.cam-name")))
+        self.K = K
+        self.D = D
+        self.xy = xy
+        return self
+
+    @classmethod
+    def build_with_multi_feature_libs(cls, imgps, feature_libs, **kwargs):
+        self = cls(**kwargs)
+        if not isinstance(feature_libs, dict):
+            feature_libs = dict([(str(i), v) for i, v in enumerate(feature_libs)])
+        kwargs["save_feature_vis"] = False
+        for name in feature_libs:
+            kwargs["name"] = self.name + "~cam_~" + name
+            feature_lib = feature_libs[name]
+            cam_ = cls(imgps, feature_lib=feature_lib, **kwargs)
+            for k in cam_:
+                cam_[k]["feature_lib"] = feature_lib
+                self[k + "~" + name] = cam_[k]
+        self.calibrate()
+        return self
+
 
 class Cams(list):
     """
