@@ -96,7 +96,7 @@ class Stereo:
             100,
             1e-5,
         )
-        ret, _, _, _, _, self.R, self.t, E, F = cv2.stereoCalibrate(
+        retval, _, _, _, _, self.R, self.t, E, F = cv2.stereoCalibrate(
             conjoint_object_points,
             conjoint_image_points1,
             conjoint_image_points2,
@@ -105,10 +105,10 @@ class Stereo:
             self.cam2.K,
             self.cam2.D,
             self.cam1.xy,
-            criteria=stereocalib_criteria,
             flags=flags,
+            criteria=stereocalib_criteria,
         )
-        self.ret = ret
+        self.retval = retval
 
     def _get_undistort_rectify_map(self):
         xy = self.cam1.xy
@@ -137,7 +137,7 @@ class Stereo:
         )
         return [rectified1, rectified2]
 
-    DUMP_ATTRS = ["R", "t"]
+    DUMP_ATTRS = ["R", "t", "retval"]
 
     def dump(self, path="", return_dict=False):
 
@@ -171,15 +171,25 @@ class Stereo:
                 with open(path_or_str) as f:
                     dic = yaml.safe_load(f)
         else:
-            dic = path_or_str_or_dict
+            dic = dict(path_or_str_or_dict)
+
+        if hasattr(self, "cam1"):
+            self.cam1.load(dic.pop("cam1"))
+            self.cam2.load(dic.pop("cam2"))
+        else:
+            self.cam1 = Cam.load(dic.pop("cam1"))
+            self.cam2 = Cam.load(dic.pop("cam2"))
         self.__dict__.update(
             {k: np.array(v) if k in self.DUMP_ATTRS else v for k, v in dic.items()}
         )
-        self.cam1 = Cam.load(self.cam1)
-        self.cam2 = Cam.load(self.cam2)
         self._get_undistort_rectify_map()
         return self
 
+    def copy(self):
+        new = type(self)()
+        new.load(self.dump())
+        return new
+
     @staticmethod
     def _get_img(path_or_np):
         if isinstance(path_or_np, str):
@@ -224,11 +234,14 @@ class Stereo:
         zs = baseline * K[0, 0] / disps
         disp_on_z = baseline * K[0, 0] / zs
 
-        print("Depth - disparity 关系:")
+        print("Depth(m) - disparity(pixel):")
         plt.plot(zs, disp_on_z)
         plt.grid()
         plt.show()
-        print("Depth - 单位Disparity深度范围关系(深度不确定度):")
+        print("\nTheoretical zmap precision:")
+        print(
+            "\tDepth - The depth range represented by a pixel disparity(depth uncertainty)"
+        )
         plt.plot(zs, np.abs(uncen_on_disp))
         # plt.grid()
         # plt.show()
@@ -236,17 +249,30 @@ class Stereo:
         # plt.plot(zs, zs / K[0, 0])
         plt.grid()
         plt.show()
-        # TODO
-        print("TODO: x,y 双目共同视野 - depth 关系:")
+        # TODO: x,y 双目共同视野 - depth 关系
         print(
             "-" * 15, "End of Stereo.precision_analysis", "-" * 15,
         )
 
     def __str__(self):
-        strr = "Stereo:\n"
-        strr += "\tBaseline: %.2fcm\n" % (100 * self.baseline)
-        strr += "\tcam1.xy: %s\n" % str(self.cam1.xy)
-        strr += "\tcam1.fovs: %s" % str(utils._str_angle_dic(self.cam1.fovs))
+        r = utils.T_to_r(self.R).squeeze()
+        du = np.linalg.norm(r) * 180 / np.pi
+        try:
+            strr = "Stereo(cam1='%s', cam2='%s'):\n" % (self.cam1.name, self.cam2.name)
+            strr += "\t xy: %s\n" % str(list(self.cam1.xy))[1:-1]
+            strr += "\t baseline: %.2fcm\n" % (100 * self.baseline)
+            strr += "\t t(cm): [%s]\n" % (
+                " ".join([str(i) for i in (self.t.squeeze() * 100).round(2)])
+            )
+            strr += "\t r(rodrigues): [%s] %.2f°\n" % (
+                " ".join([str(i) for i in r.round(3)]),
+                du,
+            )
+            strr += "\t cam1.fovs: %s\n" % str(utils._str_angle_dic(self.cam1.fovs))
+            if hasattr(self, "retval"):
+                strr += "\t retval: %s\n" % self.retval
+        except Exception as e:
+            strr += f"\t Exception({e}) in Stereo.__str__()"
         return strr
 
     __repr__ = __str__
@@ -292,17 +318,12 @@ class Stereo:
         return depth
 
     def unrectify_depth(self, depth):
-        maps = cv2.initUndistortRectifyMap(
-            self.cam1.K,
-            None,
-            np.linalg.inv(self.R1),
-            self.cam1.K,
-            self.cam1.xy,
-            cv2.CV_32FC1,
-        )
-
-        ir = cv2.remap(depth, maps[0], maps[1], cv2.INTER_NEAREST)
-        return ir
+        maps = getattr(self, "_unrectify_depth_maps", None)
+        re = utils.rotate_depth_by_remap(self.cam1.K, self.R1.T, depth, maps=maps)
+        if maps is None:
+            self._unrectify_depth_maps = re["maps"]
+        rotated_depth = re["depth"]
+        return rotated_depth
 
     def undistort_img(self, img1):
         return cv2.undistort(img1, self.cam1.K, self.cam1.D)
@@ -458,11 +479,12 @@ class SemiGlobalBlockMatching(MetaStereoMatching):
     def __call__(self, img1, img2):
         resize_ratio = min(self.max_size / max(img1.shape[:2]), 1)
         simg1, simg2 = boxx.resize(img1, resize_ratio), boxx.resize(img2, resize_ratio)
-        sdisparity = (self.stereo_sgbm.compute(simg1, simg2).astype(np.float32)).clip(
-            0
-        ) / 16.0
+        sdisparity = (self.stereo_sgbm.compute(simg1, simg2).astype(np.float32)).clip(0)
+        sdisparity[sdisparity < self.stereo_sgbm.getMinDisparity() * 16] = 0
         disparity = (
-            boxx.resize(sdisparity, img1.shape[:2]) * img1.shape[1] / simg1.shape[1]
+            boxx.resize(sdisparity / 16.0, img1.shape[:2])
+            * img1.shape[1]
+            / simg1.shape[1]
         )
         return disparity
 
@@ -494,3 +516,4 @@ if __name__ == "__main__":
     re = stereo.get_depth(img1, img2)
     boxx.tree(re)
     utils.vis_align(re["undistort_img1"], re["unrectify_depth"])
+    stereo.precision_analysis()
