@@ -39,6 +39,30 @@ def apply_T_to_point_cloud(T, point_cloud):
     return (T @ new_point_cloud.T).T[:, :3]
 
 
+def rotate_depth_by_point_cloud(K, R, depth, interpolation_rate=1):
+    pc = depth_to_point_cloud(depth, K, interpolation_rate=interpolation_rate)
+    T = np.eye(4)
+    T[:3, :3] = R
+    new_pc = apply_T_to_point_cloud(T, pc)
+    new_depth = point_cloud_to_depth(new_pc, K, depth.shape[::-1])
+    return dict(depth=new_depth, R=R)
+
+
+def rotate_depth_by_remap(K, R, depth, maps=None):
+    y, x = depth.shape
+    if maps is None:
+        maps = cv2.initUndistortRectifyMap(K, None, R, K, (x, y), cv2.CV_32FC1,)
+    ys, xs = np.mgrid[:y, :x]
+    ys, xs = ys.flatten(), xs.flatten()
+    points = np.array([xs, ys, np.ones_like(xs)]) * depth.flatten()[None]
+    # points.shape is (3, n)
+    new_points = (R @ np.linalg.inv(K) @ points).T
+    new_depth_on_old_xy = new_points[:, 2].reshape(y, x)
+
+    new_depth = cv2.remap(new_depth_on_old_xy, maps[0], maps[1], cv2.INTER_NEAREST)
+    return dict(depth=new_depth, R=R, maps=maps)
+
+
 def depth_to_point_cloud(depth, K, interpolation_rate=1):
     # depth to point cloud
     # K@Ps/z=xy
@@ -69,18 +93,21 @@ def depth_to_point_cloud(depth, K, interpolation_rate=1):
 def point_cloud_to_depth(points, K, xy):
     xyzs = (K @ points.T).T
     xyzs[:, :2] /= xyzs[:, 2:]
-
-    xyzs = np.array(sorted(xyzs, key=lambda xyz: -xyz[2]))
+    sorted_idx = np.argsort(-xyzs[:, 2])
+    xyzs = xyzs[sorted_idx]
     return xyzs_to_arr2d(xyzs, xy[::-1])
 
 
-def xyzs_to_arr2d(xyzs, hw=None, bg_value=0):
-    if hw is None:
-        hw = np.int32(xyzs.max(0)[:2].round()) + 1
+def xyzs_to_arr2d(xyzs, hw=None, bg_value=0, arr2d=None):
+    if arr2d is None:
+        if hw is None:
+            hw = np.int32(xyzs.max(0)[:2].round()) + 1
+        arr2d = np.ones(hw, xyzs.dtype) * bg_value
+    else:
+        hw = arr2d.shape[:2]
 
     xs, ys = np.int32(xyzs[:, :2].round()).T
     mask = (xs >= 0) & (xs < hw[1]) & (ys >= 0) & (ys < hw[0])
-    arr2d = np.ones(hw, xyzs.dtype) * bg_value
     arr2d[ys[mask], xs[mask]] = xyzs[:, 2][mask]
     return arr2d
 
@@ -212,7 +239,7 @@ def vis_T(T, cam=None, img=None, length=0.1):
     else:
         vis = img.copy()
     rvec, tvec = T_to_r_t(T)
-    cv2.aruco.drawAxis(vis, cam.K, cam.D, rvec, tvec, length)
+    cv2.drawFrameAxes(vis, cam.K, cam.D, rvec, tvec, length)
     return vis
 
 
